#[macro_use]
extern crate environmental;
extern crate polkadot_state_machine;

pub use std::vec::Vec;
pub use std::rc::Rc;
pub use std::cell::RefCell;
pub use std::boxed::Box;
pub use std::mem::{size_of, transmute};

use polkadot_state_machine::Externalities;
use std::fmt;

// TODO: use the real error, not NoError.

#[derive(Debug)]
struct NoError;
impl fmt::Display for NoError {
	fn fmt(&self, f: &mut fmt::Formatter) -> fmt::Result { write!(f, "") }
}

pub struct ExternalitiesHolder<'a> {
	ext: &'a mut Externalities<Error=NoError>,
}

declare_generic_environment!(ext : ExternalitiesHolder);

pub fn storage(_key: &[u8]) -> Vec<u8> {
	ext::with(|holder| holder.ext.storage(_key).ok().map(|s| s.to_vec()))
		.unwrap_or(None)
		.unwrap_or_else(|| vec![])
}

pub fn storage_into<T: Sized>(_key: &[u8]) -> Option<T> {
	let size = size_of::<T>();

	ext::with(|holder| {
		if let Ok(value) = holder.ext.storage(_key) {
			if value.len() == size {
				unsafe {
					let mut result: T = std::mem::uninitialized();
					std::slice::from_raw_parts_mut(transmute::<*mut T, *mut u8>(&mut result), size)
						.copy_from_slice(&value);
					return Some(result);
				}
			}
		}
		None
	}).unwrap_or(None)
}

pub fn set_storage(_key: &[u8], _value: &[u8]) {
	ext::with(|holder|
		holder.ext.set_storage(_key.to_vec(), _value.to_vec())
	);
}

#[macro_export]
macro_rules! impl_stubs {
	($( $name:ident ),*) => {}
}

#[cfg(test)]
mod tests {
	use super::*;
	use std::collections::HashMap;

	#[derive(Debug, Default)]
	struct TestExternalities {
		storage: HashMap<Vec<u8>, Vec<u8>>,
	}
	impl Externalities for TestExternalities {
		type Error = NoError;

		fn storage(&self, key: &[u8]) -> Result<&[u8], NoError> {
			Ok(self.storage.get(&key.to_vec()).map_or(&[] as &[u8], Vec::as_slice))
		}

		fn set_storage(&mut self, key: Vec<u8>, value: Vec<u8>) {
			self.storage.insert(key, value);
		}
	}

	macro_rules! map {
		($( $name:expr => $value:expr ),*) => (
			vec![ $( ( $name, $value ) ),* ].into_iter().collect()
		)
	}

	#[test]
	fn storage_works() {
		let mut t = TestExternalities { storage: map![], };

		{
			let mut h = ExternalitiesHolder { ext: &mut t, };
			ext::using(&mut h, || {
				assert_eq!(storage(b"hello"), b"".to_vec());
				set_storage(b"hello", b"world");
				assert_eq!(storage(b"hello"), b"world".to_vec());
				assert_eq!(storage(b"foo"), b"".to_vec());
				set_storage(b"foo", &[1, 2, 3][..]);
				assert_eq!(storage_into::<[u8; 3]>(b"foo"), Some([1, 2, 3]));
				assert_eq!(storage_into::<[u8; 3]>(b"hello"), None);
			});
		}

		t.storage = map![b"foo".to_vec() => b"bar".to_vec()];

		{
			let mut h = ExternalitiesHolder { ext: &mut t, };
			ext::using(&mut h, || {
				assert_eq!(storage(b"hello"), b"".to_vec());
				assert_eq!(storage(b"foo"), b"bar".to_vec());
			});
		}
	}
}
